"""Contains the DrillEvent class
"""
import os
import copy
import thornpy
from . import TMPLT_ENV
from .utilities import read_TO_file, get_cdb_path, get_full_path

class DrillEvent():
    """
    Creates an object with all data necessary to write a drill event.  Once the DrillEvent is instanced ramp parameters must be defined before the DrillEvent is written to an event file.  

    Note
    ----
    After instancing the class, at least one simulation step must be added using the `add_simulation_step()` method. Also, at least one ramp must be added for each of the four drilling parameters (i.e. wob, rpm, gpm, rop) using the `add_ramp()` method.  All other parameters in the event file can be specified when the DrillEvent is instanced using kwargs or they can be set later using:
        >>> drill_event.parameters[parameter] = value 
    
    Attributes
    ----------
    parameters : dict
        Dictionary of parameters that make up an Adams Drill string and would be found in an Adams Drill String file (.str).  The keys of the dictionary are the parameter names that would be seen in the string file and the values of the dictionary are the values that would be seen in the string file.
    filename : str
        Name of the event file (.evt) in which this event is stored.  This attribute is initially empty and is populated by the `write_to_file()` method. 
    """

    _SCALAR_PARAMETERS = [
        'Units',
        'Event_Name',
        'Drive_Type',
        'Measurement_Tool',
        'Start_Depth',
        'Off_Bottom',
        'Initial_Drive_Torque',
        'Motor_Type',
        'Filter_Time_Constant',
        'Motor_Bend_Start',
        'Motor_Bend_Ramp',
        'Mud_Density',
        'Impact_Damping_Penetration',
        'Impact_Exponent',
        'MWD_Pulsing',
        'NperRev',
        'N',
        'S_threshold',
        'C_hi',
        'Plotting_4D',
        'Start_Time',
        'End_Time',
        'Plotting_Interval',
        'Start_Distance',
        'End_Distance'
    ]
    
    _DEFAULT_PARAMETER_SCALARS = {
        'Units': 'Imperial',
        'Drive_Type': 'WITH_MOTOR',
        'Measurement_Tool': 'TOS',
        'Initial_Drive_Torque': 0,
        'Motor_Type': '3D',
        'Filter_Time_Constant': 0.05,
        'Motor_Bend_Start': 1.0,
        'Motor_Bend_Ramp': 9.0,
        'Mud_Density': 75,
        'Impact_Damping_Penetration': 0.005,
        'Impact_Exponent': 1.05,
        'MWD_Pulsing': 'On',
        'NperRev': 'off',
        'N': 1,
        'S_threshold': 0.55,
        'C_hi': 1.05,
        'Plotting_4D': 'off',
        'Start_Time': 160,
        'End_Time': 200,
        'Plotting_Interval': 0.1,
        'Start_Distance': 2.0,
        'End_Distance': 100.0
    }

    _TABLE_PARAMETERS = [
        'TOP_DRIVE',
        'MOTOR',
        'PUMP_FLOW',
        'WOB',
        'ROP',
        'NPERREV',
        'DYNAMICS'
    ]

    _DEFAULT_PARAMETER_TABLES = {
        'TOP_DRIVE': ((),(),()),
        'MOTOR': ((0,), (1,), (1,)),
        'PUMP_FLOW': ((),(),()),
        'WOB': ((),(),()),
        'ROP': ((),(),()),
        'NPERREV': ((0,),(1,)),
        'DYNAMICS': ((), ())
    }

    _CDB_TABLE = 'events.tbl'
    _EXT = 'evt'

    def __init__(self, event_name, start_depth, off_bottom, **kwargs):
        """Initializes the :class:`DrillEvent` class.
        
        Parameters
        ----------
        event_name : str
            Name of the event.
        start_depth : float
            Start depth of the event in feet.
        off_bottom : float
            Starting distance from bottom.
        
        """
        self.parameters = kwargs
        self.parameters['Event_Name'] = event_name
        self.parameters['Start_Depth'] = start_depth
        self.parameters['Off_Bottom'] = off_bottom        
        
        # Apply default parameters                
        self._apply_defaults()    

        self.filename = ''

    def add_simulation_step(self, end_time, output_step_size=0.05, clear_existing=False):
        """
        Adds a dynamic simulation step.
        
        Parameters
        ----------
        end_time : float
            End time of the step in seconds.
        output_step_size : float
            Output step size of the step in seconds. (defaults is 0.05)

        """
        if clear_existing:
            self.parameters['DYNAMICS'] = [[],[]]
        self.parameters['DYNAMICS'][0].append(end_time)
        self.parameters['DYNAMICS'][1].append(output_step_size)
        self.parameters['_DYNAMICS'] = zip(*self.parameters['DYNAMICS'])
    
    def add_ramp(self, parameter, start_time, ramp_duration, delta, clear_existing=False):
        """Adds a ramp to the specified ramp parameter.
        
        Parameters
        ----------
        parameter : str
            ramp parameter, options are 'TOP_DRIVE', 'WOB', 'PUMP_FLOW', or 'ROP'
        start_time : float
            Start time of the ramp.
        ramp_duration : float
            Duration of the ramp.
        delta : float
            Delta of the ramp.  Units should be in **rpm**, **lbf**, **gpm**, or **ft/sec**.
        clear_existing : bool
            If true, existing ramps for the specified parameter will be deleted.

        """
        if clear_existing:
            self.parameters[parameter] = [[],[],[]]
        self.parameters[parameter][0].append(start_time)
        self.parameters[parameter][1].append(ramp_duration)
        self.parameters[parameter][2].append(delta)
        self.parameters['_' + parameter] = zip(*self.parameters[parameter])  

    def write_to_file(self, directory=None, filename=None, cdb=None):
        """Creates an event file from the DrillEvent object.
        
        Parameters
        ----------
        write_directory : str
            Directory in which to write the file. Defaults to current working directory.
        filename : str
            Name of the file to write.  Defaults to self.parameters['EVENT_NAME']
        cdb : str
            Name of the cdb in which to write the file.  This argument overrides the write_directory.
        
        Raises
        ------
        ValueError
            Raised if not all parameters have been defined.
            
        """
        # Raise an error if the parameters can't be validated
        if not self.validate():
            raise ValueError('The parameters could not be validated.')
        
        if directory is not None:
            # If the write_directory argument is passed
            if filename is None:
                # If the filename argument is not passed, set the filename
                # as the Event_Name in the file
                filename = self.parameters['Event_Name']
                
            else:
                # If the filename argument is passed, strip the path and the 
                # extension
                filename = os.path.split(filename)[-1].replace(f'.{self._EXT}','')
            
            # Set the filepath to the filename in the given directory
            filepath = os.path.join(directory, f'{filename}.{self._EXT}')

        elif cdb is not None:
            # If the write_directory argument is not passed, but the cdb
            # argument is

            if filename is None:
                # If the filename argument is not passed, set the filename
                # as the Event_Name in the file
                filename = self.parameters['Event_Name']
            
            else:
                # If the filename argument is passed, strip the path and the 
                # extension
                filename = os.path.split(filename)[-1].replace(f'.{self._EXT}','')
            
            # Set the filepath to the file in the cdb
            filepath = get_full_path(os.path.join(f'<{cdb}>', self._CDB_TABLE, f'{filename}.{self._EXT}'))
        
        elif filename is not None:
            # If the filename argument is given
            filepath = thornpy.utilities.convert_path(filename)
        
        else:
            # Raise an error if none of the arguments are provided
            raise ValueError('Ether directory, filename, or cdb must be passed.')  
                      
        event_template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'templates', f'template.{self._EXT}')).read())
        
        with open(filepath, 'w') as fid:
            fid.write(event_template.render(self.parameters))

        self.filename = get_cdb_path(filepath)

        return self.filename

    
    def validate(self):
        """
        Determines if all parameters have been set
        
        Returns
        -------
        bool
            `True` if all parameters have been set. Otherwise `False`.

        """
        validated = True        
        # Check that all parameters exist in the self.parameters dictionary
        for param_name in self._SCALAR_PARAMETERS:
            if param_name not in self.parameters:
                validated = False        
        
        for param_name in self._TABLE_PARAMETERS:
            if not all([elem for elem in self.parameters[param_name]]):
                validated = False
            
        return validated          

    @classmethod
    def read_from_file(cls, filename):
        """Reads an Adams Drill event (.evt) file and sets :attr:`DrillEvent.parameters` based on data in the file.
        
        Parameters
        ----------
        filename : str
            Filename of an Adams Drill event (.evt) file.

        """
        # Read the TO data into a dictionary
        tiem_orbit_data = read_TO_file(get_full_path(filename))

        # Create an empty event object
        event = cls('','','')
        
        # Extract the DrillEvent parameters from the TO dictionary        
        event._get_params_from_TO_data(tiem_orbit_data) #pylint: disable=protected-access
        
        # Set the filename attribute
        event.filename = filename

        return event

    def _apply_defaults(self):
        """
        Applies defaults from class variables

        """
        # Applies normal parameter defaults
        for scalar_parameter, value in self._DEFAULT_PARAMETER_SCALARS.items():
            if scalar_parameter not in self.parameters:
                self.parameters[scalar_parameter] = copy.copy(value)

        # Applies defaults to all ramp parameters
        for table_parameter, table in self._DEFAULT_PARAMETER_TABLES.items():
            self.parameters[table_parameter] = [list(tup) for tup in table]
            self.parameters['_' + table_parameter] = zip(*self.parameters[table_parameter])
    
    def _get_params_from_TO_data(self, tiem_orbit_data):
        """Reads the event parameters out of a dictoinary of Tiem Orbit data generated by :func:`adamspy.adripy.utilities.read_TO_file`
        
        Parameters
        ----------
        tiem_orbit_data : dict
            Dictionary of Tiem Orbit data
        
        Raises
        ------
        ValueError
            An event parameter could not be found

        """
        for param in self._SCALAR_PARAMETERS:
            # For each event parameter initialize a found flag
            found = False
        
            for block in tiem_orbit_data:
                # For each block in the TO file

                if param.lower() in tiem_orbit_data[block]:
                    # If the parameter is in this block, set the parameter and break the loop
                    self.parameters[param] = tiem_orbit_data[block][param.lower()]
                    found = True
                    break

                else:
                    # If the parameter is not in this block, find all the sub blocks 
                    # and look for the parameter inside each sub block
                    sub_blocks = [header for header in tiem_orbit_data[block] if isinstance(tiem_orbit_data[block][header], dict)]
                    
                    for sub_block in sub_blocks:
                        # For each sub_block in the block

                        if param.lower() in tiem_orbit_data[block][sub_block]:
                            # If the parameter is in the sub block, set the parameter and break the loop
                            self.parameters[param] = tiem_orbit_data[block][sub_block][param.lower()]
                            found = True
                            break
                    
                    if found:
                        break
            
            # Raise a value error if the parameter isn't found.
            if not found:
                raise ValueError(f'{param} not found!')
        
        for param in self._TABLE_PARAMETERS:
            # For each event parameter initialize a found flag
            found = False

            blocks = [header for header in tiem_orbit_data if isinstance(tiem_orbit_data[header], dict)]
        
            for block in blocks:
                # For each block in the TO file, get the sub blocks
                sub_blocks = [header for header in tiem_orbit_data[block] if isinstance(tiem_orbit_data[block][header], dict)]

                if param.upper()==block and block=='DYNAMICS':
                    # If the parameter is DYNAMICS and so is the current block
                    self.parameters[param][0] = tiem_orbit_data[block]['end_time']
                    self.parameters[param][1] = tiem_orbit_data[block]['output_step_size']
                    found = True
                    break

                elif param.upper()==block and param.upper():
                    # If the current parameter is the block, add to the parameters dictionary
                    self.parameters[param][0] = tiem_orbit_data[block]['start_time']
                    self.parameters[param][1] = tiem_orbit_data[block]['ramp_duration']
                    if len(self.parameters[param])>2:
                        last_col = [col for col in tiem_orbit_data[block] if col not in ['start_time','ramp_duration']][0]
                        self.parameters[param][2] = tiem_orbit_data[block][last_col]
                    found = True
                    break
                
                elif param.upper() in sub_blocks:
                    # If the current parameter is in this sub block, add to the parameters dictionary
                    self.parameters[param][0] = tiem_orbit_data[block][param.upper()]['start_time']
                    self.parameters[param][1] = tiem_orbit_data[block][param.upper()]['ramp_duration']
                    if len(self.parameters[param])>2:
                        list_params = [par for par in tiem_orbit_data[block][param.upper()] if isinstance(tiem_orbit_data[block][param.upper()][par], list)]
                        last_col = [col for col in list_params if col not in ['start_time','ramp_duration']][0]
                        self.parameters[param][2] = tiem_orbit_data[block][param.upper()][last_col]
                    found = True
                    break
                
            if not found:
                raise ValueError(f'{param} not found!')
            
            self.parameters['_' + param] = zip(*self.parameters[param])



"""Tests related to the DrillSolverSettings class
"""

import unittest
import os
from test import *
    
from adamspy import adripy #pylint: disable=wrong-import-position

class Test_DrillSolverSettings(unittest.TestCase):
    
    maxDiff = None
    
    @classmethod
    def setUpClass(cls):        
        # Create a configuration file for testing
        adripy.create_cfg_file(TEST_CONFIG_FILENAME, [TEST_DATABASE_PATH])
        try:            
            os.remove(os.path.join(TEST_WORKING_DIRECTORY, TEST_SOLVER_SETTINGS_NAME + '.ssf'))
        except FileNotFoundError:
            pass
    
    def test_write_to_file(self):
        """Tests that the parameters in a :class:`DrillSolverSettings: object are correctly written to a solver settings file.
        """
        # Create a DrillSolverSettings object
        solver_settings = adripy.DrillSolverSettings(TEST_SOLVER_SETTINGS_NAME)

        # Write the DrillSolverSettings object to a solver settins file
        solver_settings.write_to_file(TEST_SOLVER_SETTINGS_NAME, directory=TEST_WORKING_DIRECTORY)
        
        # Check for unexpected lines in the solver settings file
        failures = check_file_contents(solver_settings.filename, TEST_EXPECTED_SOLVER_SETTINGS_FILE_TEXT)
        
        # Delete the solver settings file
        os.remove(solver_settings.filename)

        self.assertListEqual(failures, [])

    def test_read_from_file(self):        
        """Tests that the parameters in the string are correct after a string is read from a file.
        """        
        # Create an event object
        solver_settings_file = os.path.join(f'<{TEST_DATABASE_NAME}>', 'solver_settings.tbl', TEST_SOLVER_SETTINGS_NAME + '.ssf')

        # Read new parameters into the drill string object from a file
        solver_settings_from_file = adripy.DrillSolverSettings.read_from_file(solver_settings_file)
        
        params = dict(solver_settings_from_file.parameters)
        params.pop('_Funnel')
        TEST_EXPECTED_SOLVER_SETTINGS_TO_PARAMETERS.pop('_Funnel')

        self.assertDictEqual(params, TEST_EXPECTED_SOLVER_SETTINGS_TO_PARAMETERS)

    def test_read_from_file_2(self):        
        """Tests that the parameters in the string are correct after a string is read from a file.
        """        
        # Create an event object
        solver_settings_file = os.path.join(f'<{TEST_DATABASE_NAME}>', 'solver_settings.tbl', TEST_SOLVER_SETTINGS_NAME_2 + '.ssf')

        # Read new parameters into the drill string object from a file
        solver_settings_from_file = adripy.DrillSolverSettings.read_from_file(solver_settings_file)
        
        params = dict(solver_settings_from_file.parameters)
        params.pop('_Funnel')
        TEST_EXPECTED_SOLVER_SETTINGS_TO_PARAMETERS_2.pop('_Funnel')

        self.assertDictEqual(params, TEST_EXPECTED_SOLVER_SETTINGS_TO_PARAMETERS_2)

    def test_read_from_file_then_write_to_file(self):        
        """Tests that the parameters in the string are correct after a string is read from a file.
        """        
        # Create an event object
        solver_settings_file = os.path.join(f'<{TEST_DATABASE_NAME}>', 'solver_settings.tbl', TEST_SOLVER_SETTINGS_NAME + '.ssf')

        # Read new parameters into the drill string object from a file
        solver_settings_from_file = adripy.DrillSolverSettings.read_from_file(solver_settings_file)

        # Write the DrillSolverSettings object to a solver settins file
        solver_settings_from_file.write_to_file(TEST_SOLVER_SETTINGS_NAME, directory=TEST_WORKING_DIRECTORY)
        
        # Check for unexpected lines in the solver settings file
        failures = check_file_contents(solver_settings_from_file.filename, TEST_EXPECTED_SOLVER_SETTINGS_FILE_TEXT_READ_THEN_WRITE)
        
        # Delete the solver settings file
        os.remove(solver_settings_from_file.filename)

        self.assertListEqual(failures, [])

    @classmethod
    def tearDownClass(cls):
        # Delete test config file
        os.remove(TEST_CONFIG_FILENAME)
        os.environ['ADRILL_USER_CFG'] = os.path.join(os.environ['USERPROFILE'], '.adrill.cfg')
        


import xml.etree.ElementTree as et
import os
SHRUNK_RES_SUFFIX = '_shrunk'
XML_REF = 'http://www.mscsoftware.com/:xrf10'

def get_results(result_file, reqs_to_get=None, t_min=None, t_max=None, return_units=False):
	"""Gets results from an Adams results (.res) file.
	
	Example
	-------
	>>> result_file = 'example.res'
	>>> t_min = 70
	>>> t_max = 80
	>>> reqs_to_get = {}
	>>> reqs_to_get['MSE'] = ['Instantaneous_Bottom_MSE', 'Filtered_Surface_MSE']
	>>> reqs_to_get['ROP_controls'] = ['Command_ROP', 'True_WOB']
	>>> requests, units = get_results(result_file, reqs_to_get, t_min, t_max)

	Note
	----
	This funciton only works with Requests.  It does not work with Result Sets.

	Note
	----
	This function only works with xml results files.
	
	Parameters
	----------
	result_file : str
		Filename of an Adams results (.res) file
	reqs_to_get : dict
		Dictionary of requests to extract (the default is None, which gets all results)
	t_min : float, optional
		Minimum time for which to extract results (the default is None)
	t_max : float, optional
		Maximum time for which to extract results (the default is None)
	
	Returns
	-------
	dict
		Dictionary of request data
	dict
		Dictionary defining units for each request. NOTE: This is only returned if `return_units=False`
	
	"""
	# Parse the results file
	res_tree = et.parse(result_file)
	
	# Loop over all the *Entity* nodes in the input tree, pick out the ones requested
	# in `reqs_to_keep` and put their units and original ids into dictionaries
	units, req_ids, reqs_to_get = _get_units_and_ids(res_tree, reqs_to_get)

	# Initialize the output requests dictionary
	reqs_to_return = {req : {req_comp : [] for req_comp in reqs_to_get[req]} for req in reqs_to_get}
	time = []

	# Make a list of all the data nodes that represent dynamic sims
	dyn_data_nodes = [node for node in res_tree.iter('{'+XML_REF+'}Data') if 'dynamic_' in node.attrib.get('name')]

	for data_node in dyn_data_nodes:
		# For each dynamic data node

		for step_node in data_node:
			# For each step (only one) in the model input data, put the
			# old step data into a list
			step_data = step_node.text.replace('\n',' ').split(' ')
			
			t_step = round(float(step_data[1]),3)
			if (t_min is None or t_step >= t_min) and (t_max is None or t_step <= t_max):
				# If the time falls within the desired time range, add to the time list
				time.append(t_step)

				# Add to the req_comp list
				for req in reqs_to_return:
					for req_comp in reqs_to_return[req]:
						req_id = int(req_ids[req][req_comp])
						reqs_to_return[req][req_comp].append(float(step_data[req_id]))
	
	# Add the time list to the return dict
	reqs_to_return['time'] = time

	if return_units:
		return reqs_to_return, units
	else:
		return reqs_to_return


def shrink_results(result_file, reqs_to_keep=None, t_min=None, t_max=None, new_result_file=None, in_place=False):	
	"""Shrinks a results file by eliminating unwanted data.
	
	Example
	-------
	>>> result_file = 'example.res'
	>>> t_min = 70
	>>> t_max = 80
	>>> reqs_to_keep = {}
	>>> reqs_to_keep['MSE'] = ['Instantaneous_Bottom_MSE', 'Filtered_Surface_MSE']
	>>> reqs_to_keep['ROP_controls'] = ['Command_ROP', 'True_WOB']
	>>> shrink_results(result_file, reqs_to_keep, t_min, t_max)

	Note
	----
	This funciton only works with Requests.  It does not work with Result Sets.

	Parameters
	----------
	result_file : str
		Filename of the results file
	reqs_to_keep : dict
		Dictionary of requests and request components that should be saved. 
	t_min : float, optional
		Start time of period of interest (default is None which uses the first time step)
	t_max : float, optional
		End time of period of interest (default is None which uses the last time step)
	new_result_file : str
		Name that will be given to the new results file. (Default is None which adds a '_shrunk' suffix to `result_file` if in_place=False or uses `result_file` if in_place=True.)
	in_place : bool, optional
		If True, existing file will be deleted.

	"""	
	input_tree = et.parse(result_file)
	input_ans_node = [ans for ans in input_tree.iter('{'+XML_REF+'}Analysis')][0]
	
	# Create `units` and `old_ids` dictionaries
	units, old_ids, _reqs_to_keep = _get_units_and_ids(input_tree, reqs_to_keep)
	
	# If `reqs_to_keep` is None, set it equal to all requrests
	if reqs_to_keep is None:
		reqs_to_keep = _reqs_to_keep

	# Create New XML tree
	root_res = et.Element('Results', xmlns=XML_REF)    
	root_ans = et.SubElement(root_res, 'Analysis', name=input_ans_node.attrib['name'], executionDate=input_ans_node.attrib['executionDate'], Solver=input_ans_node.attrib['Solver'], script=input_ans_node.attrib['script'])
	root_map = et.SubElement(root_ans, 'StepMap', name='map_001')
	ent_time = et.SubElement(root_map, 'Entity', name='time')
	root_data = et.SubElement(root_ans, 'Data', name='dynamic_001', id='1')
				
	# Create the time component
	et.SubElement(ent_time, 'Component', name='TIME', unitsValue='sec', id='1')

	# Create StepMap
	new_id = 2
	for request in reqs_to_keep:
		ent_map = et.SubElement(root_map, 'Entity', name=request, entity=request, entType='Request', objectId='0')
		for req_comp in reqs_to_keep[request]:
			et.SubElement(ent_map, 'Component', name=req_comp, plotLabel=units[request][req_comp], id=str(new_id))
			new_id += 1

	# Make a list of all the data nodes that represent dynamic sims
	dyn_data_nodes = [node for node in input_tree.iter('{'+XML_REF+'}Data') if 'dynamic_' in node.attrib.get('name')]

	# Loop through the dynamic sim data nodes
	for data_node in dyn_data_nodes:
		# For each dynamic data node

		for step_node in data_node:
			# For each step (only one) in the model input data, put the
			# old step data into a list
			old_step_data = step_node.text.replace('\n',' ').split(' ')

			t_step = round(float(old_step_data[1]),3)
			if (t_min is None or t_step >= t_min) and (t_max is None or t_step <= t_max):
				# If the time falls within the desired time range add to `new_step_data`
				new_step_data = _get_new_steplist(reqs_to_keep, old_step_data, old_ids)

				# Create a string containing the model input data								
				new_dynamic_text = _convert_steplist_to_string(new_step_data)
			
				# Rewrite the SubElement text
				xml_step = et.SubElement(root_data, 'Step', type='dynamic')
				xml_step.text = new_dynamic_text

	# Determine the output file if it wasn't given or if the one that was provided is equal to the original
	if new_result_file is None or new_result_file == result_file:	
		new_result_file = result_file + '.tmp' if in_place is True else result_file.replace('.res', '') + f'{SHRUNK_RES_SUFFIX}.res'
	
	# Write the output file
	output_tree = et.ElementTree(root_res)
	_xml_format(root_res)
	output_tree.write(new_result_file, encoding='utf-8', xml_declaration=True)

	# If in_place is set to true, remove the original results file
	if in_place is True:
		os.remove(result_file)
		
		# If no name was given for the output file and a .tmp extension was used, remove it
		if new_result_file.endswith('.tmp'):
			os.rename(new_result_file, new_result_file.replace('.tmp',''))

def _get_new_steplist(reqs_to_keep, old_step_data, req_ids):
	"""Returns a list similar to `old_step_data` but with unwanted requests removed.

	Uses the requests and request components in `old_step_data` and the entitiy ids in `req_ids` to determine which elements in `old_step_data` to keep.
	
	Parameters
	----------
	reqs_to_keep : dict
		Dictionary of requests and request components to keep
	old_step_data : list
		List of all the step data in the results file
	req_ids : dict
		Dictionary of entity ids for the entire results file
	
	Returns
	-------
	list
		List of just the step data to keep
	
	"""
	# Start a new list with just the time element from `old_step_data`  
	new_step_data = [old_step_data[1]]

	# Loop through the desired requests and components to pull elements
	# from `old_step_data` into `new_step_data`	
	for request in reqs_to_keep:
		# For each desired request 
		for req_comp in reqs_to_keep[request]:
			# For each desired request component, add that components
			# step data to `new_step_data`
			req_id = int(req_ids[request][req_comp])
			new_step_data.append(old_step_data[req_id])
	
	return new_step_data

def _convert_steplist_to_string(step_data):
	"""Converts list of step data into a single string.
	
	Parameters
	----------
	step_data : list
		List of step data
	
	Returns
	-------
	str
		A space delimited string where every 6th value is followed by a newline.
	
	"""
	text = ''

	for i, datum in enumerate(step_data):
		if i == 0:
			text += f'\n{datum}\n'
		else:
			if i%6 == 0:
				text += f'{datum}\n'
			else:
				text += f'{datum} '
	
	return text


def _xml_format(element, level=0):
	"""Makes the root text look prettier.
	
	Parameters
	----------
	elem : Element
		:class:`Element` whose text will be formatted.
	level : int, optional
		Tree level of `Element` (the default is 0)
	
	"""
	i = '\n' + level*"  "
	if len(element):
		if not element.text or not element.text.strip():
			element.text = i + "  "
		if not element.tail or not element.tail.strip():
			element.tail = i
		for elem in element:
			_xml_format(elem, level+1)
		if not element.tail or not element.tail.strip():
			element.tail = i
	else:
		if level and (not element.tail or not element.tail.strip()):
			element.tail = i

def _get_units_and_ids(tree, reqs=None):
	"""Loops over all the *Entity* nodes in `tree` and picks out the ones requested	in `reqs`. Returns dictionaries of their units and original ids.
	
	Parameters
	----------
	tree : ElementTree
		Root xml tree for an Adams results file
	reqs : dict, optional
		Dictionary of requests and request components to keep. (Default is None, which gets all the requests)
	
	Returns
	-------
	dict
		Two level dictionary indicating the units associated with the request components
	dict
		Two level dictionary indicating the xml ids associated with the request components
	dict
		Two level dictionary indicating the requests and request components NOTE: This is identical to the `reqs` argument unless the reqs argument is None.)
	"""
	units = {}
	req_ids = {}
	requests = {}

	# Make a list of the request entities in `tree`
	req_ents = [ent for ent in tree.iter('{'+XML_REF+'}Entity') if 'entType' in ent.attrib and ent.attrib['entType'] == 'Request']
	
	# Loop through all the request entities
	for entity in req_ents:
		# For each entitiy in the results tree

		if reqs is None or entity.attrib['name'] in reqs:
			# If this entity is in `reqs_to_get` or `reqs_to_get` is None
			# Store the request name in `request` 
			request = entity.attrib['name']

			# Create an entry in the `units` and `req_ids` dictionaries
			units[request] = {}
			req_ids[request] = {}
			requests[request] = {}

			for comp in entity:
				# For each component in the entitiy
				# Store the component name as `req_comp`
				req_comp = comp.attrib['name']

				if reqs is None or req_comp in reqs[request]:
					# If this component is in `reqs_to_get[request]`
					# Create an entry in the `units` and `req_ids` dictionaries 
					units[request][req_comp] = comp.attrib['plotLabel']
					req_ids[request][req_comp] = comp.attrib['id']
					requests[request][req_comp] = comp.attrib['id']
					
	return units, req_ids, requests
	


"""Tests related to using :mod:`adampsy.adripy.utilities` to read tiem orbit files.
"""

import unittest
import os
from test import *
    
from adamspy import adripy #pylint: disable=wrong-import-position
from adamspy.adripy import utilities #pylint: disable=wrong-import-position
from adamspy.adripy.constants import TO_PARAMETER_PATTERN

class Test_ReadTOFile(unittest.TestCase):
    """Tests adamspy.adripy.utilities.read_to_file()    
    """
    maxDiff = None
    def setUp(self):
        adripy.create_cfg_file(TEST_CONFIG_FILENAME, [TEST_DATABASE_PATH, TEST_NEW_DATABASE_PATH])

    def test_full_path_works(self):
        """Tests if :meth:`read_TO_file` works with a full path is passed.
        """
        try:
            _parameters = utilities.read_TO_file(adripy.get_full_path(os.path.join(f'<{TEST_DATABASE_NAME}>', 'drill_strings.tbl', TEST_EXISTING_STRING_NAME + '.str')))
        except FileNotFoundError:
            self.fail('File not found when full path used')
    
    def test_cdb_alias_works(self):
        """Tests if :meth:`read_TO_file` works with a cdb alias path is passed.
        """
        try:
            _parameters = utilities.read_TO_file(os.path.join(f'<{TEST_DATABASE_NAME}>', 'drill_strings.tbl', TEST_EXISTING_STRING_NAME + '.str'))
        except FileNotFoundError:
            self.fail('File not found when cdb alias path used')
        
    def test_correct_values(self):
        """Tests that the parameters in the written TO file are as expected
        """
        # Create a test config file containing the test database
        parameters = utilities.read_TO_file(adripy.get_full_path(os.path.join(f'<{TEST_DATABASE_NAME}>', 'drill_strings.tbl', TEST_EXISTING_STRING_NAME + '.str')))

        self.assertDictEqual(parameters, EXPECTED_STRING_TO_PARAMETERS)    

    def tearDown(self):
        try:
            os.remove(TEST_CONFIG_FILENAME)
        except FileNotFoundError:
            pass
        os.environ['ADRILL_USER_CFG'] = os.path.join(os.environ['USERPROFILE'], '.adrill.cfg')

class Test_RexExPatterns(unittest.TestCase):
    """Tests the regular expressions in adripy.utilities    
    """
    def setUp(self):
        return

    def test_block_header_pattern(self):
        strings_to_match = [
            '[LKJFDSLKJ]\n',
            '[lkjfdslkj]\n',
            '[LKJFDLKJ_FDSD]\n',
            '[lkjfdlkjfd_lkjfdslkjfds]\n',
            '[PLOT_4D]\n'
        ]
        strings_to_not_match = [
            '$[LKJFDLKJ]\n',
            '(LKJFDSLKJ)\n'
        ]
        failures = self._find_match_failures(strings_to_match, strings_to_not_match, utilities.TO_BLOCK_HEADER_PATTERN)        
        self.assertEqual(failures, [])

    def test_subblock_header_pattern(self):
        strings_to_match = [
            '(LKJFDSLKJ)\n',
            '(lkjfdslkj)\n',
            '(LKJFDLKJ_FDSD)\n',
            '(lkjfdlkjfd_lkjfdslkjfds)\n',
            '(PLOT_4D)\n'
        ]
        strings_to_not_match = [
            '[LKJFDLKJ]\n',
            '$(LKJFDSLKJ)\n'
        ]
        failures = self._find_match_failures(strings_to_match, strings_to_not_match, utilities.TO_SUBBLOCK_HEADER_PATTERN)
        self.assertEqual(failures, [])

    def test_table_header_pattern(self):
        strings_to_match = [
            '{lkjfdslkj lkjfds fdslkjs}\n',
            '{FDSSDF lkjfds fdslkjs}\n',
            '{FDSLKJFDSLKJ LKJFDLKJ FDSLKJFDS}\n',
            '{FDSLKJ_FDSLKJ LKJFDLKJ FDSLKJFDS}\n',
            '{ lkjfdslkj lkjfds fdslkjs }\n'
        ]
        strings_to_not_match = [
            '{FDSLKJ-FDSLKJ, LKJFDLKJ FDSLKJFDS}\n',
            '[DSF]\n',
            '${lkjfdslk lkjfdslkj}\n'
        ]             
        failures = self._find_match_failures(strings_to_match, strings_to_not_match, utilities.TO_TABLE_HEADER_PATTERN)      
        self.assertEqual(failures, [])

    def test_table_line_pattern(self):
        strings_to_match = [
            ' 432 32.3 -43 -43.45 \n',
            '432 32.3 -43 -43.45\n',
            '432  32.3  -43  -43.45 \n',
            '432  5.55e+02  -4e-3  -43.45 \n',
            '234.3\n',
            '\'lkjfdslkj\'\n',
            '\'lkjfdslkj\' \'DLKDSFLK\'\n',
            '  \'lkjfdslkj\' \'DLKDSFLK\'  \n',
            '\'lkjfdslkj\' \'DLKDSFLK_4D\'\n'            
        ]
        strings_to_not_match = [
            '$ 432 32.3 -43 -43.45 \n',
            '$ [SDFSDFS]\n',
            '$ (SDFSDFS)\n'
        ]             
        failures = self._find_match_failures(strings_to_match, strings_to_not_match, utilities.TO_TABLE_LINE_PATTERN)      
        self.assertEqual(failures, [])

    def test_parameter_pattern(self):
        strings_to_match = [
            ' lkjfdslkj  =  \'lkjfdslkj\' \n',
            ' lkjfdslkj  =  \'lkjfdslkj\'\n',
            ' lkjfdslkj  =  100\n',
            ' lkjfdslkj  =  100.00\n',
            ' lkjfdslkj  =  -100\n',
            ' lkjfdslkj  =  -100.00\n',
            ' DSFDSFFDD  =  100\n',
            ' DSFdffdsf  =  100\n',
            ' Plotting_4D		= \'on\'',
            ' Hmax =   5.0e-3',
            ' PP =   5.0e+3',
            '  Pipe_Length         = 31',
            " Event_Property_File  =  ''",
            '  Stabilizer_Length= 3.0000'
        ]
        strings_to_not_match = [
            'lkjfds  =  2\n',
            '$ lkjfds  =  2\n',
            ' fdslkj  =  "lkjfdslkj"\n',
            ' fdslkfds  =  5*10'
        ]             
        failures = self._find_match_failures(strings_to_match, strings_to_not_match, TO_PARAMETER_PATTERN)      
        self.assertEqual(failures, [])

    def _find_match_failures(self, strings_to_match, strings_to_not_match, pattern):
        """Test that strings_to_match matches pattern and strings_to_not_match 
        does not match pattern.
        
        Parameters
        ----------
        strings_to_match : str
            Strings that should match pattern
        strings_to_not_match : str
            Strings that should not match pattern
        pattern : re.match
            Match object
        
        Returns
        -------
        list
            List of failures
        """

        failures = []
        # Add missed strings to failures list
        for string in strings_to_match:
            if not pattern.match(string):
                failures.append(f'{string} erroneously not matched')
        
        # Add erroneously matched strings to failures list
        for string in strings_to_not_match:
            if pattern.match(string):
                failures.append(f'{string} erroneously matched')
        
        return failures

    def tearDown(self):
        return

if __name__ == '__main__':
    unittest.main()



"""Tests related to the DrillEvent class
"""

import unittest
import os
from test import *
    
from adamspy import adripy #pylint: disable=wrong-import-position

class Test_EventFile(unittest.TestCase):
    """
    Tests that adripy can correctly write an event file
    """
    maxDiff = None
    def setUp(self):
        
        # Create a configuration file for testing
        adripy.create_cfg_file(TEST_CONFIG_FILENAME, [TEST_DATABASE_PATH])

        # Create event file object
        self.event = adripy.DrillEvent(TEST_CREATED_EVENT_NAME, 4000, 4)
        
        # Add ramp parameters to event file object
        self.event.add_ramp('PUMP_FLOW', 1, 10, 500, clear_existing=True)
        self.event.add_ramp('TOP_DRIVE', 10, 15, 60, clear_existing=True)
        self.event.add_ramp('WOB', 30, 10, 50, clear_existing=True)
        self.event.add_ramp('ROP', 35, 10, 100, clear_existing=True)   

        # Add simulation steps to event file object
        self.event.add_simulation_step(10, clear_existing=True)
        self.event.add_simulation_step(100)     

        # Write an event file from the event file object
        self.event.write_to_file(cdb=TEST_DATABASE_NAME)

    def test_validate_with_an_empty_table(self):
        """Tests if :meth:`DrillEvent.validate` returns False when one of the table parameters is not set.        
        """
        event = adripy.DrillEvent(TEST_CREATED_EVENT_NAME, 4000, 4)
        event.add_ramp('PUMP_FLOW', 1, 10, 500, clear_existing=True)
        event.add_ramp('TOP_DRIVE', 10, 15, 60, clear_existing=True)
        event.add_ramp('WOB', 30, 10, 50, clear_existing=True)
        
        validated = event.validate()
        self.assertFalse(validated)

    def test_read_from_file(self):
        """Tests that the parameters in the string are correct after a string is read from a file.
        """
        
        # Create an event object
        event_file = os.path.join(f'<{TEST_DATABASE_NAME}>', 'events.tbl', TEST_EVENT_NAME + '.evt')

        # Read new parameters into the drill string object from a file
        event_from_file = adripy.DrillEvent.read_from_file(event_file)
        
        params = dict(event_from_file.parameters)
        params.pop('_DYNAMICS')
        params.pop('_MOTOR')
        params.pop('_NPERREV')
        params.pop('_PUMP_FLOW')
        params.pop('_ROP')
        params.pop('_TOP_DRIVE')
        params.pop('_WOB')

        self.assertDictEqual(params, TEST_EXPECTED_EVENT_TO_PARAMETERS)
    
    def test_write_event_file(self):
        """
        Tests if the event file was written correctly.
        """
        # Check content of created event file
        failures = check_file_contents(TEST_CREATED_EVENT_FILENAME, EXPECTED_EVENT_FILE_TEXT)        
            
        self.assertListEqual([], failures)

    def tearDown(self):

        # Delete test config file
        os.remove(TEST_CONFIG_FILENAME)
        os.environ['ADRILL_USER_CFG'] = os.path.join(os.environ['USERPROFILE'], '.adrill.cfg')

        # Delete the test event file
        os.remove(TEST_CREATED_EVENT_FILENAME)

if __name__ == '__main__':
    unittest.main()



"""Contains the DrillTool class
"""

import os
import sys
import traceback
import shutil
from thornpy import numtype
from .utilities import TO_BLOCK_HEADER_PATTERN, get_cdb_location, get_cdb_path, get_full_path
from .constants import DATABASE_INFO, TO_PARAMETER_PATTERN

class DrillTool():
    """
    Object representing an Adams Drill tool.  Instances must be generated from a tool property file.
    
    Attributes
    ----------
    property_file : str
        Tiem Orbit file representing the drill tool.
    name : str
        Name of the tools
    tool_type : str
        Type of the tool.
    extension : str
        Extension used for tiem orbit file
    table : str
        Name of the cdb table used for the particular tool type

    """     
    _ACCEPTED_TYPES = [
        'drillpipe',
        'drill_collar',
        'accelerator',
        'stabilizer',
        'short_collar',
        'dart',
        'jar',
        'agitator',
        'blade_reamer',
        'crossover',
        'darts',
        'flex_pipe',
        'hw_pipe',
        'pdc_bit',
        'motor',
        'shock_sub',
        'lwd_tool',
        'mfr_tool',
        'mwd_tool',
        'instrumentation_sub',
        'generic_long',
        'generic_short',
        'top_drive',
        'equivalent_upper_string'
    ]

    def __init__(self, property_file):
        """Initializes the `DrillTool` object.
        
        Parameters
        ----------
        property_file : str
            Tiem Orbit file representing the drill tool.

        """ 
        self._check_extension(property_file)
        self.property_file = get_cdb_path(property_file)
        self.name = self._get_name()
        self.tool_type = self._get_type()    
        self.extension = self._get_extension()
        self.table = self._get_table()        
        self._name_receivers = []

        
    
    def bind_name_to(self, name_receiver):
        """Appends `name_receiver` to :attr:`_name_receivers`.
        
        All the methods in the `name_receiver` :obj:list will be called anytime the :meth:`rename()` method is called

        Parameters
        ----------
        name_receiver : func
            Any method that takes a :class:`DrillTool`.

        """
        self._name_receivers.append(name_receiver)
    
    def rename(self, new_name, remove_original=False):
        """Renames the tool
        
        Note
        ----
        If this tool is part of a :class:`DrillString` object, the :class:`DrillString` object will be updated with the new name and property file.

        Parameters
        ----------
        new_name : str
            New name for the tool.
        remove_original : bool
            If `True` will rename the property file.  If `False` will create a new one.

        """
        # Determine the new filename
        current_filename = get_full_path(self.property_file)
        current_filepath = os.path.split(current_filename)[0]
        new_filename = os.path.join(current_filepath, f'{new_name}.{self.extension}')
        
        # Copy the file to the new location
        shutil.copyfile(current_filename, new_filename)
        self.property_file = get_cdb_path(new_filename)   

        # Modify the property file
        self.modify_parameter_value('Name', new_name)

        # Change the name variable
        self.name = new_name     
        
        if remove_original is True:
            # Delete the old property file
            os.remove(current_filename)    
                
        # NOTE: The for loop below updates the name and property file in the DrillString object
        # Run all the name receivers 
        for name_receiver in self._name_receivers:
            name_receiver(self)

    def copy_file(self, directory=None, cdb=None):
        """Copies the file to `directory` or to the appropriate table in `cdb`.
        
        Parameters
        ----------
        write_directory : str
            Directory in which to write the file. (default is None.)            
        cdb : str
            Name of the cdb in which to write the file.  This argument overrides the directory. (default is None.)
                    
        Raises
        ------
        ValueError
            Raised if neither directory nor cdb are given.
        ValueError
            Raised if not all parameters have been defined.
        
        Note
        ----
        Either `directory` or `cdb` must be given. 

        """
        # Check that directory or cdb was given.
        if directory is None and cdb is None:
            raise ValueError('Either directory or cdb is required!')
        
        # Determine the filename for writing
        if cdb is not None:
            # If cdb was given
            filename = os.path.join(get_cdb_location(cdb), self.table, f'{self.name}.{self.extension}')            
        else:
            # If cdb not given and directory was given
            filename = os.path.join(directory, f'{self.name}.{self.extension}')

        # Copy the file to the new location
        shutil.copyfile(get_full_path(self.property_file), filename)

        # Change the property_file attribute
        self.property_file = get_cdb_path(filename)
    
    def get_parameter_value(self, parameter_to_get):
        """
        Returns the value of the specified parameter from the property file.  
        
        Note
        ----
        This method cannot get parameters from the UNITS block.
        
        Parameters
        ----------
        parameter_to_get : str
            Name of the parameter to get from the property file.
        
        Returns
        -------
        :obj:`string` or :obj:`float`
            Value of the specified parameter from the property file.

        """
        # Initilize return variable as None
        found = False

        # Read the property file
        filename = get_full_path(self.property_file)
        with open(filename, 'rb') as fid:            
            lines_b = fid.readlines()

        # Decode each line ignore any exceptions that occur on commented lines
        lines = []
        for line_b in lines_b:
            try:
                lines.append(line_b.decode())

            except UnicodeDecodeError as err:
                if line_b[0:1].decode() == '$':
                    pass
                else:
                    raise err

        # Initialize a boolean specifying if the for loop is at a line in the [UNITS] block
        at_units_block = False
        
        for line in lines:                
            # For each line in the property file
            
            if at_units_block and TO_BLOCK_HEADER_PATTERN.match(line):
                # If end of units block reached 
                at_units_block = False

            elif line.startswith('[UNITS]'):
                # If beginning of units block reached
                at_units_block = True

            if TO_PARAMETER_PATTERN.match(line) and at_units_block is False:                
                # If the line matches the pattern of a parameter definition

                # Split the line into parameter and value
                [current_parameter, current_value] = line.replace('\n','').replace(' ','').split('=')
                if current_parameter.lower() == parameter_to_get.lower():
                    # If the parameter on the current line is the parameter to get
                    
                    if "'" in current_value:
                        # If the value is a string
                        value = current_value.replace("'",'').strip()
                    
                    else:
                        # If the value is a number
                        value = float(current_value) if numtype.str_is_float(current_value) else int(current_value)
                    found = True
                    break
        
        if not found:
            # If the parameter wasn't found raise an exception
            filename = self.property_file
            raise DrillToolError(f'The parameter {parameter_to_get} was not found in {filename}!')

        return value

    def modify_parameter_value(self, parameter_to_change, new_value):
        """Modifies a parameter in a :class:`DrillTool` property file.
        
        Parameters
        ----------
        parameter_to_change : str
            Name of the parameter to modify.  Must match the name in the property file. (Not case sensitive). 
        new_value : :obj:`float` or :obj:`str`
            New value of the parameter to change
        new_filename : str
            Filename of new property file.  If None the original property file will be overwritten (the default is None.)

        """
        filename = get_full_path(self.property_file)        
            
        # Read the lines in the original file into a list
        with open(filename, 'r') as fid:
            original_lines = fid.readlines()

        # Write a new list with the correct line modified
        new_lines = []
        for line in original_lines:                
            # For each line in the original file, append to new_lines
            new_lines.append(line)

            if TO_PARAMETER_PATTERN.match(line):                
                # If the line matches the pattern of a parameter definition
                [current_parameter, value] = line.replace('\n','').replace(' ','').split('=')

                if current_parameter.lower() == parameter_to_change.lower():
                    # If the parameter is the parameter to be changed

                    if "'" in value:
                        # If the value is a string, add quotes
                        new_value = "'{}'".format(new_value)
                    
                    new_lines[-1] = ' {}  =  {}\n'.format(current_parameter, new_value)

        with open(filename, 'w') as fid:
            fid.writelines(new_lines)

    def _get_name(self):
        return self.get_parameter_value('Name')
    
    def _get_type(self):
        return self.get_parameter_value('File_Type').lower()
    
    def _get_extension(self):
        extension = DATABASE_INFO[self.tool_type]['extension']
        return extension

    def _get_table(self):
        table = DATABASE_INFO[self.tool_type]['table']
        return table

    @classmethod
    def _check_extension(cls, property_file):      
        """Checks if the extension on `property_file` is in the list of supported tool file types.
        
        Parameters
        ----------
        property_file : str
            Filename of an Adams Drill property file.
        
        Raises
        ------
        DrillToolError
            Raised if the extension on `property_file` is not in the list of supported tool file types.
        
        """
        ext = os.path.splitext(property_file)[1]
        if ext not in ['.' + info['extension'] for name, info in DATABASE_INFO.items() if name in cls._ACCEPTED_TYPES]:
            raise DrillToolError(f'The extension {ext} is not supported by the DrillTool class.')

class DrillToolError(Exception):
    pass